from .http.request import Request
from .http.response import Response
from .routing.router import Router
import uvicorn
from asyncio import iscoroutinefunction
import inspect


class Ushka:
    def __init__(self, app_path=None) -> None:
        # Get the path of the file that instantiated Ushka
        # stack()[1] refers to the caller's frame
        # .f_globals['__file__'] gets the __file__ attribute from that frame
        if app_path is None:
            app_path = inspect.stack()[1].frame.f_globals["__file__"]

        self.app_path = app_path
        self.router = Router(app_path)
        self.router.autodiscover()

    async def handle_http_request(self, scope, receive, send):
        request = Request(scope, receive)

        func, params = self.router.get_route(request)

        if callable(func):
            if iscoroutinefunction(func):
                result = await func(**params)
            else:
                result = func(**params)

            # check result type to auto detect media type (example html, text, json)

            if "<html>" in result:
                response = Response(result, media_type="text/html")
            else:
                response = Response(result)
        else:
            response = Response("Not Found", 404)

        await response(send)

    async def handle_lifespan(self, receive, send):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                await send({"type": "lifespan.startup.complete"})
            elif message["type"] == "lifespan.shutdown":
                await send({"type": "lifespan.shutdown.complete"})
            return

    async def handle_asgi_call(self, scope, receive, send):
        if scope["type"] == "http":
            await self.handle_http_request(scope, receive, send)

        elif scope["type"] == "lifespan":
            await self.handle_lifespan(receive, send)

        else:
            response = Response("Not Supported", 501)

            await response(send)

    async def __call__(self, scope, receive, send):
        await self.handle_asgi_call(scope, receive, send)

    def run(self, host="127.0.0.1", port=8000):
        print(f"Starting Ushka server on http://{host}:{port}")
        uvicorn.run(self, host=host, port=port)



class Response:
    def __init__(self, content, status_code=200, media_type="text/plain") -> None:
        self.content = content

        self.status_code = status_code

        self.media_type = media_type

    async def __call__(self, send):
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": [[b"content-type", self.media_type.encode()]],
            }
        )

        await send({"type": "http.response.body", "body": self.content.encode()})



import re
from typing import Dict, Callable, Tuple

from ..http.request import Request


def normalize_url_path(path: str) -> str:
    parts = [p for p in path.split("/") if p and not p == "."]
    return "/" + "/".join(parts)


class Router:
    def __init__(self, app_path: str):
        self.app_path = app_path
        self.static_routes: Dict[str, Dict[str, Callable]] = {}
        self.dynamic_routes: Dict[
            str, list[Tuple[re.Pattern, list[str], Callable]]
        ] = {}

    def add_route(self, method: str, path: str, func: Callable):
        path = normalize_url_path(path)

        if "[" not in path:
            self.static_routes.setdefault(method, {})[path] = func

        else:
            param_names = []
            regex_pattern = "^"
            for part in path.strip("/").split("/"):
                if part.startswith("[") and part.endswith("]"):
                    param = part[1:-1]
                    param_names.append(param)
                    regex_pattern += r"/(?P<%s>[^/]+)" % param
                else:
                    regex_pattern += "/" + part

            regex_pattern += "$"
            compiled = re.compile(regex_pattern)
            self.dynamic_routes.setdefault(method, []).append(
                (compiled, param_names, func)
            )

    def get_route(self, request: Request) -> Tuple[Callable, Dict] | Tuple[None, Dict]:
        method = request.method
        path = normalize_url_path(request.path)

        func = self.static_routes.get(method, {}).get(path)
        if func:
            return func, {}

        for regex, param_names, dynamic_func in self.dynamic_routes.get(method, []):
            # TODO: Add Resolve Route to injects depends like func(request)
            match = regex.match(path)
            if match:
                return dynamic_func, match.groupdict()

        return None, {}

    def autodiscover(self, folder="routes"):
        from pathlib import Path
        import importlib.util
        import inspect

        base_path = Path(self.app_path).parent
        target_path = base_path.joinpath(folder)
        for file in target_path.rglob("*.py"):
            if not file.name.startswith("__"):
                spec = importlib.util.spec_from_file_location("mod", str(file))
                if spec is None or spec.loader is None:
                    continue

                mod = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(mod)

                for name, obj in inspect.getmembers(mod):
                    if inspect.isfunction(obj):
                        # TODO: Add ANY method mode
                        if name.upper() in [
                            "GET",
                            "POST",
                            "PUT",
                            "UPDATE",
                            "DELETE",
                            "HEAD",
                        ]:
                            # FIXME: Race condition caused by /foo/index.py and /foo.py It doesn't pose a danger, but it can cause overwriting and make a "bug" difficult to spot.

                            if file.name.lower() == "index.py":
                                relative_path = file.parent.relative_to(target_path)

                            else:
                                relative_path = str(
                                    file.relative_to(target_path)
                                ).removesuffix(".py")

                            self.add_route(name.upper(), str(relative_path), obj)
